################################################################################
#
# Copyright (C) 2016-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import os

import pandas as pd
import numpy as np
import yaml
import subprocess
import glob

from shutil import copyfile
from copy import deepcopy

from .Utilities.ConditionalImports import yamlLoader
from . import LibraryIO

from . import ClientWriter
from .Common import assignGlobalParameters, ensurePath, globalParameters, \
    gfxName, gfxArch, printExit, getArchitectureName
from .SolutionStructs import ProblemSizes


def getArchitecture(isaName):
    archid = gfxName(isaName)
    return getArchitectureName(archid)

def isValidArch(archName, currentArch):
    arch = gfxArch(archName)
    return currentArch == arch

##############################################################################
# createLibraryForBenchmark
##############################################################################
def createLibraryForBenchmark(logicPath, libraryPath, currentPath):
    """
    takes the path of existing logic files as input and adds the summation
    model for each of the solutions. This is used in the Tile Aware Metirc
    Selection.
    """

    pythonExePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), "bin", "TensileCreateLibrary")
    args = [pythonExePath, \
        "--merge-files", "--new-client-only", "--no-short-file-names", \
        "--architecture=all", "--cxx-compiler="+globalParameters["CxxCompiler"], "--library-format=yaml", \
        logicPath, libraryPath, "HIP"]

    try:
        subprocess.run(args, check=True, cwd=currentPath)
    except (subprocess.CalledProcessError, OSError) as e:
        printExit("ClientWriter Benchmark Process exited with error: {}".format(e))

def GenerateSummations(userArgs):

    inputLogicPath = userArgs[0]
    outputPath = userArgs[1]
    assignGlobalParameters({})

    currentISA = globalParameters["CurrentISA"]
    currentArchitecture = getArchitecture(currentISA)

    globPath = os.path.join(inputLogicPath, "{}*".format(currentArchitecture))
    logicFileNames = glob.glob(globPath)

    for logicFileName in logicFileNames:

        logicFileBaseName = os.path.basename(logicFileName)
        logicFileStem, ext = os.path.splitext(logicFileBaseName)
        if (ext != ".yaml"):
            continue

        currentPath = ensurePath(os.path.join(outputPath, logicFileStem))

        libPath = ensurePath(os.path.join(currentPath, "lib"))
        finalPath = ensurePath(os.path.join(currentPath, "final"))
        localLogicPath = ensurePath(os.path.join(currentPath, "logic"))
        localLogicFilePath = os.path.join(localLogicPath, logicFileBaseName)

        # Here we read in two version of the logic the first one fills the solutions with
        # defaults and modifies some of the parameters. The final logic file should be the
        # same as the initial logic with the summation model added. To preseve the original
        # logic we also read in the raw unaltered version of the logic and stage the content
        # to write the final logic.
        logic    = LibraryIO.parseLibraryLogicFile(logicFileName)
        rawLogic = LibraryIO.rawLibraryLogic(logicFileName)

        # If we cannot read the logic file then skip it
        if rawLogic == None or logic == None:
            printExit("Error reading the file: %s. skipping." % logicFileName)

        (versionStringR, scheduleNameR, architectureNameR, deviceNamesR, problemTypeStateR,\
            solutionStatesR, indexOrderR, exactLogicR, rangeLogicR, otherFieldsR) =\
            rawLogic

        copyfile(logicFileName, localLogicFilePath)
        createLibraryForBenchmark(localLogicPath, libPath, currentPath)

        exactList = []

        solutionSummationSizes = [32,64,96,128,256,512,1024,2048,4096,8192,16384]

        for K in solutionSummationSizes:
            e = {"Exact" : [8192, 4096, 1, K]}
            exactList.append(e)

        libraryPath = libPath
        clientBuildDir = os.path.join(outputPath, "client")
        problemTypeObj = logic.problemType.state

        problemSizes = ProblemSizes(problemTypeObj, exactList)
        dataPath = ensurePath(os.path.join(outputPath, logicFileStem, "data"))
        configFilePath = ensurePath(os.path.join(outputPath, logicFileStem, "configs"))
        dataFilePath = os.path.join(dataPath, "benchmark.csv")
        configFile = os.path.join(configFilePath, "ClientParameters.ini")
        scriptPath = ensurePath(os.path.join(outputPath, logicFileStem, "script"))

        ClientWriter.CreateBenchmarkClientParametersForSizes(libraryPath, problemSizes, dataFilePath, configFile, problemTypeObj)
        ClientWriter.runNewClient(scriptPath, configFile, clientBuildDir)

        tensileLibraryFile = os.path.join(libPath, "library", "TensileLibrary.yaml")

        stream = open(tensileLibraryFile, "r")
        tensileLibrary = yaml.load(stream, yamlLoader)
        stream.close()

        libSolutions = tensileLibrary["solutions"]

        libSolutionNames = []
        for s in libSolutions:
            kernelName=s["name"]
            libSolutionNames.append(kernelName)

        working_data=pd.read_csv(dataFilePath).rename(str.strip,axis='columns')

        index_keys = working_data.SizeL.unique()
        solutionsDF = working_data.filter(like='Cij')

        perf_max = solutionsDF.max().max().item()

        solutionIndex = 0
        for s_stateR, kernelName in zip(solutionStatesR, libSolutionNames):
            solutionIndex += 1
            perf_raw = working_data[kernelName]
            perf = (1000*index_keys) / perf_raw
            model = np.polyfit(x=index_keys, y=perf, deg=1)
            slope = model[0].item()
            intercept = model[1].item()
            linearModel = {"slope": slope, "intercept": intercept, "max": perf_max}
            s_stateR["LinearModel"] = deepcopy(linearModel)

        rawLogicData = []
        rawLogicData.append(deepcopy(versionStringR))
        rawLogicData.append(deepcopy(scheduleNameR))
        rawLogicData.append(deepcopy(architectureNameR))
        rawLogicData.append(deepcopy(deviceNamesR))
        rawLogicData.append(deepcopy(problemTypeStateR))
        rawLogicData.append(deepcopy(solutionStatesR))
        rawLogicData.append(deepcopy(indexOrderR))
        rawLogicData.append(deepcopy(exactLogicR))
        rawLogicData.append(deepcopy(rangeLogicR))
        for idx in range(0, len(otherFieldsR)):
            rawLogicData.append(deepcopy(otherFieldsR[idx]))

        localFinalLogic = os.path.join(finalPath, logicFileBaseName)

        yamlWriter = LibraryIO.YAMLWriter()
        yamlWriter.write(localFinalLogic, rawLogicData)

        outputFinal = ensurePath(os.path.join(outputPath, "final"))
        finalLogic = os.path.join(outputFinal, logicFileBaseName)

        copyfile(localFinalLogic, finalLogic)



################################################################################
#
# Copyright (C) 2020-2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import pytest
import itertools
import copy
from Tensile.Hardware import HardwarePredicate
from Tensile.SolutionLibrary import PredicateLibrary


def test_hardware_predicate_comparison():
    a = HardwarePredicate.FromISA((9,0,0))
    b = HardwarePredicate.FromISA((9,0,6))
    c = HardwarePredicate("TruePred")
    d = HardwarePredicate.FromHardware((9,0,8), 60)
    e = HardwarePredicate.FromHardware((9,0,8), 64)
    f = HardwarePredicate.FromHardware((9,4,2))
    g = HardwarePredicate.FromHardware((9,4,2), isAPU=0)
    h = HardwarePredicate.FromHardware((9,4,2), isAPU=1)

    assert a < b
    assert a < c
    assert b < c

    assert d < a
    assert d < b
    assert d < c

    assert e < a
    assert e < b
    assert e < c
    assert e < d

    assert g < a
    assert g < b
    assert g < c
    assert g < d
    assert g < e
    assert g < f

    assert h < a
    assert h < b
    assert h < c
    assert h < d
    assert h < e
    assert h < f
    assert h < g

    assert not b < a
    assert not c < a
    assert not c < b
    assert not a < e
    assert not b < e
    assert not c < e
    assert not d < e
    assert not f < g
    assert not g < h

    assert not a < a
    assert not b < b
    assert not c < c
    assert not d < d
    assert not e < e
    assert not f < f
    assert not g < g
    assert not h < h

def hardware_library_objects_order():
    objs = [PredicateLibrary('Hardware', [{'predicate': HardwarePredicate.FromISA((9,0,0))}]),
            PredicateLibrary('Hardware', [{'predicate': HardwarePredicate.FromISA((9,0,6))}]),
            PredicateLibrary('Hardware', [{'predicate': HardwarePredicate.FromISA((9,0,8))}]),
            PredicateLibrary('Hardware', [{'predicate': HardwarePredicate.FromISA((9,0,10))}]),
            PredicateLibrary('Hardware', [{'predicate': HardwarePredicate.FromHardware((9,0,8), 60)}]),
            PredicateLibrary('Hardware', [{'predicate': HardwarePredicate.FromHardware((9,0,8), 64)}]),
            PredicateLibrary('Hardware', [{'predicate': HardwarePredicate('TruePred')}])
    ]

    return [copy.deepcopy(libs) for libs in itertools.permutations(objs)]

@pytest.mark.parametrize("libraries", hardware_library_objects_order())
def test_hardware_library_merge_order(libraries):
    lib = libraries[0]
    for lib2 in libraries[1:]:
        lib.merge(lib2)

    assert lib.rows[-1]['predicate'] == HardwarePredicate('TruePred')
    assert lib.rows[0]['predicate'] == HardwarePredicate.FromHardware((9,0,8), 64)
    assert lib.rows[1]['predicate'] == HardwarePredicate.FromHardware((9,0,8), 60)
    for r in lib.rows[:-1]:
        assert r['predicate'] != HardwarePredicate('TruePred')

def hardware_library_objects_order2():
    objs = [PredicateLibrary('Hardware', [{'predicate': HardwarePredicate.FromISA((9,0,6))}]),
            PredicateLibrary('Hardware', [{'predicate': HardwarePredicate.FromHardware((9,4,2))}]),
            PredicateLibrary('Hardware', [{'predicate': HardwarePredicate.FromHardware((9,4,2), isAPU=0)}]),
            PredicateLibrary('Hardware', [{'predicate': HardwarePredicate.FromHardware((9,4,2), isAPU=1)}]),
            PredicateLibrary('Hardware', [{'predicate': HardwarePredicate('TruePred')}])
    ]

    return [copy.deepcopy(libs) for libs in itertools.permutations(objs)]

@pytest.mark.parametrize("libraries", hardware_library_objects_order2())
def test_hardware_library_merge_order2(libraries):
    lib = libraries[0]
    for lib2 in libraries[1:]:
        lib.merge(lib2)

    assert lib.rows[-1]['predicate'] == HardwarePredicate('TruePred')
    assert lib.rows[0]['predicate'] == HardwarePredicate.FromHardware((9,4,2), isAPU=1)
    assert lib.rows[1]['predicate'] == HardwarePredicate.FromHardware((9,4,2), isAPU=0)
    assert lib.rows[3]['predicate'] == HardwarePredicate.FromHardware((9,4,2))
    for r in lib.rows[:-1]:
        assert r['predicate'] != HardwarePredicate('TruePred')

def hardware_library_objects_dups():
    objs = [PredicateLibrary('Hardware', [{'predicate': HardwarePredicate.FromISA((9,0,0)), 'library': PredicateLibrary()}]),
            PredicateLibrary('Hardware', [{'predicate': HardwarePredicate.FromISA((9,0,6)), 'library': PredicateLibrary()}]),
            PredicateLibrary('Hardware', [{'predicate': HardwarePredicate.FromISA((9,0,6)), 'library': PredicateLibrary()}]),
            PredicateLibrary('Hardware', [{'predicate': HardwarePredicate.FromISA((9,0,8)), 'library': PredicateLibrary()}]),
            PredicateLibrary('Hardware', [{'predicate': HardwarePredicate('TruePred'),      'library': PredicateLibrary()}])
    ]

    return [copy.deepcopy(libs) for libs in itertools.permutations(objs)]

@pytest.mark.parametrize("libraries", hardware_library_objects_dups())
def test_hardware_library_merge_dups(libraries):
    lib = libraries[0]
    for lib2 in libraries[1:]:
        lib.merge(lib2)

    assert len(lib.rows) == 4

    def getPred(row):
        return row['predicate']
    rowPreds = map(getPred, lib.rows)

    assert HardwarePredicate.FromISA((9,0,0)) in rowPreds
    assert HardwarePredicate.FromISA((9,0,6)) in rowPreds
    assert HardwarePredicate.FromISA((9,0,8)) in rowPreds
    assert HardwarePredicate('TruePred') in rowPreds



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from ..Component import Component, MAC
from ..DataType import DataType

class MAC_F32_Plain(MAC):
    """
    Plain MAC instruction implementation
    """
    @staticmethod
    def asmCaps(caps):
        return caps["v_mac_f32"] or caps["v_fma_f32"]
    #archCaps = {}
    kernel = {"ProblemType": {"DataType": DataType(DataType.single)}}

    def __call__(self, writer, m, innerUnroll):
        kernel = writer.kernel

        instruction = "v_mac_f32"
        if kernel["MACInstruction"] == "FMA":
            if writer.asmCaps["v_fmac_f32"]:
                instruction = "v_fmac_f32"
            elif writer.asmCaps["v_fma_f32"]:
                instruction = "v_fma_f32"
            else:
                raise RuntimeError("FMA instruction specified but not supported on {}".format(kernel["ISA"]))

        if not writer.asmCaps[instruction]:
            raise RuntimeError("{} instruction specified but not supported on {}".format(instruction, kernel["ISA"]))

        kStr = self.commentHeader()

        vars = {}

        vars["m"] = m
        vars["kernel"] = kernel
        vars["endLine"] = writer.endLine

        vars["ThreadTile0"] = kernel["ThreadTile0"]
        vars["ThreadTile1"] = kernel["ThreadTile1"]
        vars["PerformanceWaitCount"] = kernel["PerformanceWaitCount"]

        vars["instruction"] = instruction

        priority = Component.Priority.find(writer)
        macIdx = 0

        for iui in range(0, innerUnroll):
            for idx1 in range(0, kernel["ThreadTile1"]):
                for idx0 in range(0, kernel["ThreadTile0"]):
                    vars["idx0"] = idx0
                    vars["idx1"] = idx1
                    vars["a"] = idx0 if writer.tPB["tile01Idx"] else idx1
                    vars["b"] = idx1 if writer.tPB["tile01Idx"] else idx0
                    vars["iui"] = iui

                    vars["cStr"] = "v[vgprValuC + {idx0} + {idx1}*{ThreadTile0}]".format_map(vars)
                    vars["aStr"] = "v[vgprValuA_X{m}_I{iui} + {a}]".format_map(vars)
                    vars["bStr"] = "v[vgprValuB_X{m}_I{iui} + {b}]".format_map(vars)

                    if instruction == "v_fma_f32":
                        kStr += "v_fma_f32 {cStr}, {aStr}, {bStr}, {cStr}{endLine}".format_map(vars)
                    else:
                        kStr += "{instruction} {cStr}, {aStr}, {bStr}{endLine}".format_map(vars)

                    kStr += priority(writer, 1, "Raise priority while processing macs")

                    if macIdx == kernel["PerformanceWaitLocation"]:
                        kStr += "s_waitcnt lgkmcnt({PerformanceWaitCount}) // extra wait for performance{endLine}".format_map(vars)
                    if macIdx == kernel["PerformanceSyncLocation"]:
                        kStr += "s_barrier // extra barrier for performance{endLine}".format_map(vars)
                    macIdx += 1

        kStr += priority(writer, 0, "Reset priority after macs")

        return kStr



################################################################################
#
# Copyright (C) 2019-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import pytest
import subprocess
from collections import namedtuple

from YamlBuilder.YamlBuilder import YamlBuilder
from YamlBuilder.YamlBuilder import Solutions

TestConfig=namedtuple("TestConfig", ["solution", "problem_func"])
Runner=namedtuple("Runner", ["level", "func", "solution"])
TensileState=namedtuple("TensileState", ["args"])

# this control the default solutions used for each test.
solutions = ["src1", "src5_gsu", "asm3_pbd", "asm3_splitu", "asm3_mi"]
#solutions = ["asm3_pbd"] # run a smaller set of tests

@pytest.fixture(scope="function")
def file_with_test_name(request, tmp_path):
    def get(suffix):
        filename = request.node.name + suffix
        return tmp_path.joinpath(filename)
    return get

@pytest.fixture
def run_nothing():
    def run(conv, problemType, solution=Solutions.defaultSolution(), problemFunc=None, problemLevel=-1, dataType='s'):
        pass
    return run

@pytest.fixture
def run_generate_yaml(request, file_with_test_name):
    def run(conv, problemType, solution=Solutions.defaultSolution(), problemFunc=None, problemLevel=-1, dataType='s'):
        if problemFunc == None:
            problemFunc = YamlBuilder.ProblemSizes
        if problemLevel==-1:
            problemLevel = request.config.getoption("--problem-level")
        config = YamlBuilder.ConvolutionContraction(conv, problemType, solution, dataType, problemFunc, True, problemLevel)
        configFile = file_with_test_name(".contraction.yaml")
        print("Generate_YAML output:", configFile)
        config.write(configFile)
        return configFile
    return run

@pytest.fixture
def run_contraction(tensile_args, tmp_path, run_generate_yaml, request, tensile_script_path):
    def run(conv, problemType, solution, problemFunc=None, problemLevel=-1, dataType='s'):
        configFile = run_generate_yaml(conv, problemType, solution, problemFunc, problemLevel, dataType)
        args = [str(configFile), str(tmp_path), *tensile_args]
        # change to use  check_output to force windows cmd block util command finish
        subprocess.check_output([tensile_script_path] + args, stderr=subprocess.STDOUT)
    return run

@pytest.fixture
def run_convolution_vs_contraction(tensile_args, tmp_path, file_with_test_name, tensile_script_path):
    def run(conv, problemType={}, solution=Solutions.src1, dataType='s'):
        config = YamlBuilder.ConvolutionVsContraction(conv, solution, dataType)
        configFile = file_with_test_name(".conv.yaml")
        config.write(configFile)

        args = [str(configFile), str(tmp_path), *tensile_args]
        # change to use  check_output to force windows cmd block util command finish
        subprocess.check_output([tensile_script_path] + args, stderr=subprocess.STDOUT)
    return run

level_params = [pytest.param((0, Solutions.src1), id="Convolution_Class"),
                pytest.param((1, Solutions.defaultSolution()), id="Generate_YAML-" + Solutions.defaultSolution().__name__)] + \
               [pytest.param((2, getattr(Solutions,s)), id="Run_Contraction-"+s) for s in solutions]
               #[pytest.param((3, Solutions.defaultSolution()), id="Run_Convolution_vs_Contraction:" + Solutions.defaultSolution().__name__)]

@pytest.fixture(params=level_params)
def run_convolution_level(request,
                          pytestconfig,
                          run_nothing,
                          run_generate_yaml,
                          run_contraction,
                          run_convolution_vs_contraction):
    level_fixtures = [run_nothing,
                      run_generate_yaml,
                      run_contraction,
                      run_convolution_vs_contraction]
    curLevel = request.param[0]
    argLevel = request.config.getoption("--test-level")
    if curLevel > argLevel:
        pytest.skip()
    return Runner(curLevel, level_fixtures[curLevel], request.param[1])

def pytest_addoption(parser):
    parser.addoption(
        "--no-conv-assertions", action="store_true", default=1,
        help='''
        Disable assertiong on Convolution class.
        ''')

    parser.addoption(
        "--test-level", action="store", type=int, default=2,
        help='''
        0= test Tensile.Convolution class;
        1= 0 plus generate YAML;
        2= 1 plus run tensile client and compare generated contraction vs CPU;
        3= 2 plus run tensile_client with convolution-vs-contraction (only forward conv tests which define Spatial will PASS )
        '''
        )
    parser.addoption(
        "--problem-level", action="store", type=int, default=2,
        help='''
        How many exact configurations to generate for contraction testing for non-src solutions (typically asm).
        1= single problem
        2= tens of problems
        3= hundreds of problems
        '''
        )

@pytest.fixture()
def tensile_state(request):
    """
    Shared tensile state, including args.
    """
    args={}
    try:
        args["no_conv_assertions"]= request.config.getoption('--no-conv-assertions')
    except ValueError:
        args["no_conv_assertions"]=1
    return TensileState(args=args)



################################################################################
#
# Copyright (C) 2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from pathlib import Path
from unittest.mock import MagicMock, call, patch

import pytest

from Tensile.TensileCreateLib.KernelFileContext import (
    KernelFileContextManager,
    _closeKernelFiles,
    _openKernelFiles,
)


@pytest.fixture
def mock_openFile():
    mockFile = MagicMock()
    mockOpen = MagicMock(return_value=mockFile)
    with patch("builtins.open", mockOpen):
        yield mockOpen, mockFile


@pytest.fixture
def mock_pathResolve():
    with patch("pathlib.Path.resolve", return_value=Path("/fake/path")) as mockResolve:
        yield mockResolve


def test_openKernelFiles_withOnlyMergeFilesEnabled(mock_openFile, mock_pathResolve):
    x, y = _openKernelFiles(0, True, False, Path("/some/path"), [])
    mock_openFile[0].assert_called()
    assert mock_openFile[0].call_args_list == [
        call(Path("/fake/path/Kernels.cpp"), "a", encoding="utf-8"),
        call(Path("/fake/path/Kernels.h"), "a", encoding="utf-8"),
    ]


def test_openKernelFiles_withOnlyLazyLoadingEnabled(mock_openFile, mock_pathResolve):
    x, y = _openKernelFiles(0, False, True, Path("/some/path"), [])
    mock_openFile[0].assert_called()
    assert mock_openFile[0].call_args_list == [
        call(Path("/fake/path/Kernels.cpp"), "a", encoding="utf-8"),
        call(Path("/fake/path/Kernels.h"), "a", encoding="utf-8"),
    ]


def test_openKernelFiles_withMergeAndLazyEnabled(mock_openFile, mock_pathResolve):
    x, y = _openKernelFiles(2, True, True, Path("/fake/path"), [])
    mock_openFile[0].assert_called()
    assert mock_openFile[0].call_args_list == [
        call(Path("/fake/path/Kernels.cpp"), "a", encoding="utf-8"),
        call(Path("/fake/path/Kernels.h"), "a", encoding="utf-8"),
    ]


def test_openKernelFiles_withAll(mock_openFile, mock_pathResolve):
    x, y = _openKernelFiles(2, True, True, Path("/fake/path"), ["a_kernel"])
    mock_openFile[0].assert_called()
    assert mock_openFile[0].call_args_list == [
        call("a_kernel.cpp", "a", encoding="utf-8"),
        call("a_kernel.h", "a", encoding="utf-8"),
    ]


def test_openKernelFiles_withNumMergedFiles(mock_openFile, mock_pathResolve):
    x, y = _openKernelFiles(2, False, False, Path("/some/path"), ["kernel1.cpp", "kernel2.cpp"])
    assert mock_openFile[0].call_args_list == [
        call("kernel1.cpp", "a", encoding="utf-8"),
        call("kernel1.h", "a", encoding="utf-8"),
    ]


def test_openKernelFiles_noAction(mock_openFile, mock_pathResolve):
    _openKernelFiles(2, False, False, Path("/some/path"), [])
    mock_openFile[0].assert_not_called()


def test_closeKerneFiles(mock_openFile, mock_pathResolve):
    _, mock_file = mock_openFile
    _closeKernelFiles(mock_file, mock_file)
    assert mock_file.close.call_count == 2


def test_KernelFileContextManager_withOnlyLazyLoadingEnabled(mock_openFile, mock_pathResolve):
    with pytest.raises(ValueError, match="If lazy loading is enabled, merge files must be as well"):
        with KernelFileContextManager(True, False, 1, Path("/some/path"), []):
            pass


def test_KernelFileContextManager_withOnlyMergeFilesEnabled(mock_openFile, mock_pathResolve):
    with KernelFileContextManager(False, True, 1, Path("/some/path"), []):
        mock_openFile[0].assert_called()


def test_KernelFileContextManager_withNoMergeFilesAndNum(mock_openFile, mock_pathResolve):
    with KernelFileContextManager(False, False, 3, Path("/some/path"), ["a_kernel.cpp"]) as (x, y):
        assert x is None and y is None, "We expect nothing to be opened in this situation"


def test_KernelFileContextManager_withNoOutputPath(mock_openFile, mock_pathResolve):
    with pytest.raises(ValueError, match="Output path cannot be empty."):
        with KernelFileContextManager(False, False, 3, "", ["a_kernel.cpp"]):
            pass


def test_KernelFileContextManager_withNoKernelFiles(mock_openFile, mock_pathResolve):
    with pytest.raises(ValueError, match=r"(.*) Provide at least one kernel file."):
        with KernelFileContextManager(False, True, 3, Path("/fake/path"), []):
            pass



################################################################################
#
# Copyright (C) 2019-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import pytest,logging
from Tensile.SolutionStructs import Convolution,ConvolutionConfig
log =logging.getLogger("testlog")

def test_spatial_in():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                     })

    log.debug(conv.printUsage(z))
    pcc = ConvolutionConfig(spatial=[14,15])
    p = conv.makeProblem(n=64, c=1024, k=256, pcc=pcc)
    assert(p[0] == [210, 256, 64, 1024])
    assert(p[1] == [1, 210, 215040])
    assert(p[2] == [-1, -1, 0])

def test_spatial_parm():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Spatial' : '13x14',
                     })

    log.debug(conv.printUsage(z))
    pcc = ConvolutionConfig()
    p = conv.makeProblem(n=64, c=1024, k=256, pcc=pcc)
    assert(p[0] == [182, 256, 64, 1024])
    assert(p[1] == [1, 182, 186368])
    assert(p[2] == [-1, -1, 0])


def test_stride():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Spatial' : '13x14',
                      'Stride' : '2x3',
                     })

    log.debug(conv.printUsage(z))
    p = conv.makeProblem(n=64, c=1024, k=256, pcc=conv.cc)
    assert(p[0] == [5, 7, 256, 64, 1024])
    assert(p[1] == [3, 28, 182, 186368])
    assert(p[2] == [-1, -1, 0])


def test_stride_filter():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Spatial' : '13x14',
                      'Stride' : '2x3',
                      'Filter' : '3x4',
                     })

    log.debug(conv.printUsage(z))
    p = conv.makeProblem(n=64, c=1024, k=256, pcc=conv.cc)
    assert(p[0] == [4, 6, 256, 64, 3, 4, 1024])
    assert(p[1] == [1, 14, 3, 28, 182, 186368])
    assert(p[2] == [-1, -1, -1, -1, 0])

def test_stride_filter_dilated():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Spatial' : '13x14',
                      'Stride' : '2x3',
                      'Filter' : '3x4',
                      'Dilation': '2x3',
                     })

    log.debug(conv.printUsage(z))
    p = conv.makeProblem(n=64, c=1024, k=256, pcc=conv.cc)
    assert(p[0] == [2, 5, 256, 64, 3, 4, 1024])
    assert(p[1] == [3, 28, 3, 28, 182, 186368])
    assert(p[2] == [-1, -1, -1, -1, 0])

def test_spatial_unspecified():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                     })

    log.debug(conv.printUsage(z))
    with pytest.raises(RuntimeError, match="ConvolutionConfig field 'spatial' == None"):
        conv.makeProblem(n=64, c=1024, k=256, pcc=conv.cc)

def test_mismatch():
    z={} # problemType definition
    conv = Convolution(z, 'ConvolutionForward',
              config={'TensorAFormat': 'NCHW',
                      'Spatial' : '34x99',
                     })

    log.debug(conv.printUsage(z))
    pcc = ConvolutionConfig(spatial=[14,15])
    with pytest.raises(RuntimeError, match="Mismatch between ConvolutionConfig value"):
        conv.makeProblem(n=64, c=1024, k=256, pcc=pcc)
